// This unit contains routines for the (Pseudo) Random Number Generator
// based on the RC4 stream cipher

unit PRNG;

interface

type
  TCardDeck = array[1..52] of integer;  // card deck array
  PCardDeck = ^TCardDeck;

// forward declarations

procedure Prng_Shuffle(Cards: PCardDeck);

implementation
uses Windows, SHA256, SyncObjs;

var
  ii, jj: byte;
  ss: array[0..255] of byte;
  SeedHash: array[0..31] of byte;
  Prng_CS: TCriticalSection;

// RC4 key setup routine using seed created from Windows message timings

procedure Prng_Setup;
var
  n: integer;
  i, t: byte;
  k: array[0..255] of byte;
begin
  i := 0;
  for n := 0 to 255 do
  begin
    ss[n] := n;
    k[n] := SeedHash[i];
    Inc(i);
    if i > 31 then i := 0;
  end;
  i := 0;
  for n := 0 to 255 do
  begin
    i := byte(i + ss[n] + k[n]);
    t := ss[n];
    ss[n] := ss[i];
    ss[i] := t;
  end;
  ii := 0;
  jj := 0;
end;

// Get next byte from RNG

function Prng_Byte: byte;
var t: byte;
begin
  Inc(ii);
  jj := byte(jj + ss[ii]);
  t := ss[ii];
  ss[ii] := ss[jj];
  ss[jj] := t;
  t := byte(ss[ii] + ss[jj]);
  Result := ss[t];
end;

// Get 0 to Range - 1 value from RNG
// Updated in ver 5.10 to remove modulo bias

function Prng_Value(Range: word): word;
var x, Divisor: integer;
begin
  // make divisor a multiple of the range to remove modulo bias
  Divisor := (65536 div Range) * Range;
  repeat
    x := Prng_Byte;
    x := (x shl 8) or Prng_Byte; // combine 2 bytes into single 16-bit random word
  until x < Divisor;
  x := (x * Range) div Divisor;
  Result := x;
end;

// Shuffle card deck using RNG and Knuth's shuffle algorithm

procedure Prng_Shuffle(Cards: PCardDeck);
var i, j, k: integer;
begin
  Prng_CS.Enter;
  try
    for i := 52 downto 2 do
    begin
      k := Prng_Value(i) + 1;
      j := Cards[k];
      Cards[k] := Cards[i];
      Cards[i] := j;
    end;
  finally
    Prng_CS.Leave;
  end;
end;

// Init the PRNG with hashed seed built from system timer
// Updated in ver 5.10 to remove user input timings

procedure Prng_Init;
var
  i: integer;
  c: int64;
  PrngSeed: array[0..31] of byte;
  Sha: TSHAObject;
begin
  // create seed
  Randomize;
  for i := 0 to 31 do
  begin
    QueryPerformanceCounter(c);  // grab counter from 1.193 MHz system timer
    PrngSeed[i] := byte(c);      // keep only lower 8 bits
    Sleep(Random(3) + 1);        // 1 to 3 msec sleep allows lower 8 bits to cycle about 10 to 30 times
  end;
  // hash seed
  Sha := TSHAObject.Create;
  Sha.HashBuffer(@PrngSeed, 32);
  Sha.FinalizeHash;
  Move(Sha.Hash, SeedHash, 32);
  Sha.Free;
  Prng_Setup; // setup RC4
  for i := 0 to 255 do Prng_Byte; // cycle off 256 bytes due to RC4 key leakage
end;

initialization
  Prng_CS := TCriticalSection.Create;
  Prng_Init;
finalization
  Prng_CS.Free;
end.

